package de.chaubold.gleshelper;

import de.chaubold.math.Vec2;
import de.chaubold.math.Vec3;
import android.content.res.Resources;
import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.HashMap;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.Vector;

import javax.microedition.khronos.opengles.GL10;

public class ObjModel {

    private class Face {

        public int[] vertices = new int[3];
        public int[] texCoords = new int[3];
        public int[] normals = new int[3];
    }

    private class Group {

        public int start = 0;
        public int end = 0;
        public String name = null;
    }
    private boolean hasTexCoords = false;
    private Map<String, Material> materials = null;
    private String materiallib = null;
    private int numFaces = 0;
    private String filename = null;
    private VBO vbo = new VBO();
    private VBOwithTex vboWt = new VBOwithTex();


    public ObjModel() {
        materials = new HashMap<String, Material>();
    }

    private void CheckError(GL10 gl, String debug) {
        if (gl.glGetError() != gl.GL_NO_ERROR) {
            System.out.println("Error at " + debug);
        }
    }

    public boolean getHasTexCoords() {
        return (materiallib != null);
        //return hasTexCoords;
    }

    public Map<String, Material> getMaterials() {
        return materials;
    }

    public void setMaterials(Map<String, Material> mats) {
        materials = mats;
    }

    private void readMaterialLib(Resources R, int resId) {
        try {
            BufferedReader f = new BufferedReader(new InputStreamReader(R.openRawResource(resId)));
            String line = null;
            Material mat = null;

            while ((line = f.readLine()) != null) {
                StringTokenizer st = new StringTokenizer(line, " ");
                if (line.startsWith("newmtl")) {
                    st.nextToken();// skip the "newmtl"
                    if (mat != null) {
                        materials.put(mat.name, mat);
                    }
                    mat = new Material(st.nextToken());
                } else if (mat != null) {
                    if (line.startsWith("map_Kd")) {
                        st.nextToken(); //skip the map_Kd

                        mat.texture = st.nextToken();
                    } else if (line.startsWith("Kd")) {
                        st.nextToken(); //skip the Kd
                        Vec3 diffuse = new Vec3(1.0f, 1.0f, 1.0f);
                        diffuse.x = Float.valueOf(st.nextToken()).floatValue();
                        diffuse.y = Float.valueOf(st.nextToken()).floatValue();
                        diffuse.z = Float.valueOf(st.nextToken()).floatValue();

                        mat.diffuse = diffuse;
                    } else if (line.startsWith("Ks")) {
                        st.nextToken(); //skip the Ks
                        Vec3 specular = new Vec3(1.0f, 1.0f, 1.0f);
                        specular.x = Float.valueOf(st.nextToken()).floatValue();
                        specular.y = Float.valueOf(st.nextToken()).floatValue();
                        specular.z = Float.valueOf(st.nextToken()).floatValue();

                        mat.specular = specular;
                    } else if (line.startsWith("Ka")) {
                        st.nextToken(); //skip the Ka
                        Vec3 ambient = new Vec3(1.0f, 1.0f, 1.0f);
                        ambient.x = Float.valueOf(st.nextToken()).floatValue();
                        ambient.y = Float.valueOf(st.nextToken()).floatValue();
                        ambient.z = Float.valueOf(st.nextToken()).floatValue();

                        mat.ambient = ambient;
                    }
                }
            }
            if (mat != null) {
                materials.put(mat.name, mat);
            }
        } catch (FileNotFoundException e) {
            System.out.println("ObjModel: Wasnt able to open MaterialLib file: " + materiallib);
        } catch (IOException e) {
            System.out.println("ObjModel: Error when reading line");
        }
    }

    public void loadModel(Resources R, int model_id, int material_id, GL10 gl) throws FileNotFoundException {
        System.out.println("Loading Obj-Model: " + R.getResourceName(model_id));
        Vector<Vec3> vertices = new Vector<Vec3>();
        Vector<Vec3> normals = new Vector<Vec3>();
        Vector<Face> faces = new Vector<Face>();
        Vector<Vec2> texCoords = new Vector<Vec2>();
        Vector<Group> groups = new Vector<Group>();
        try {
            BufferedReader f = new BufferedReader(new InputStreamReader(R.openRawResource(model_id)));

            filename = R.getResourceName(model_id);
            String line = null;

            while ((line = f.readLine()) != null) {

                if (line.startsWith("mtllib")) {
                    StringTokenizer st = new StringTokenizer(line, " ");
                    st.nextToken();// skip the "mtllib"
                    materiallib = st.nextToken();
                    readMaterialLib(R, material_id);
                } else if (line.startsWith("usemtl")) {
                    StringTokenizer st = new StringTokenizer(line, " ");
                    st.nextToken();// skip the "usemtl"
                    Group g = new Group();
                    g.name = st.nextToken();
                    g.start = faces.size();
                    if (groups.size() > 0) {
                        groups.lastElement().end = faces.size();
                    }

                    groups.add(g);
                } else if (line.startsWith("vt")) {
                    StringTokenizer st = new StringTokenizer(line, " ");
                    st.nextToken();// skip the vt
                    Vec2 tmp = new Vec2();
                    try {
                        tmp.x = Float.valueOf(st.nextToken()).floatValue();
                        tmp.y = Float.valueOf(st.nextToken()).floatValue();
                        texCoords.add(tmp);
                    } catch (Exception e) {
                        System.out.println("OBJModel: Error: not enough texCoord data specified: " + st.countTokens());
                    }
                } else if (line.startsWith("vn")) {
                    // Read normal - has to consist of 3 floats
                    StringTokenizer st = new StringTokenizer(line, " ");
                    st.nextToken();// skip the vn
                    Vec3 tmp = new Vec3();
                    try {
                        tmp.x = Float.valueOf(st.nextToken()).floatValue();
                        tmp.y = Float.valueOf(st.nextToken()).floatValue();
                        tmp.z = Float.valueOf(st.nextToken()).floatValue();
                        normals.add(tmp);
                    } catch (Exception e) {
                        System.out.println("OBJModel: Error: not enough normal data specified: " + st.countTokens());
                    }
                } else if (line.startsWith("v")) {
                    // Read vertex - has to consist of 3 floats
                    StringTokenizer st = new StringTokenizer(line, " ");
                    st.nextToken();// skip the v
                    Vec3 tmp = new Vec3();
                    try {
                        tmp.x = Float.valueOf(st.nextToken()).floatValue();
                        tmp.y = Float.valueOf(st.nextToken()).floatValue();
                        tmp.z = Float.valueOf(st.nextToken()).floatValue();
                        vertices.add(tmp);
                    } catch (Exception e) {
                        System.out.println("OBJModel: Error: not enough vertex data specified: " + st.countTokens());
                    }
                } else if (line.startsWith("f")) {
                    // read face
                    StringTokenizer st = new StringTokenizer(line, " ");
                    if (st.countTokens() != 4) {
                        System.out.println("OBJModel: Can only load triangulated models!");
                        System.exit(1);
                    }
                    st.nextToken();// skip the f

                    //If we're not in a group yet, we'll fix that now
                    if (groups.size() == 0) {
                        groups.add(new Group());
                        groups.lastElement().name = "default";
                        groups.lastElement().start = 0;
                    }


                    Face temp = new Face();
                    for (int i = 0; i < 3; i++) {
                        StringTokenizer st2 = new StringTokenizer(st.nextToken(), "/");
                        if (st2.countTokens() < 1) {
                            System.out.println("OBJModel: No tokens to read..");
                            System.exit(1);
                        } else if (st2.countTokens() < 3) {
                            if (hasTexCoords) {
                                System.out.println("OBJModel: TexCoords need to be specified throughout!");
                                System.exit(1);
                            }
                            temp.vertices[i] = Integer.parseInt(st2.nextToken()) - 1;
                        } else if (st2.countTokens() == 3) {
                            hasTexCoords = true;
                            temp.vertices[i] = Integer.parseInt(st2.nextToken()) - 1;
                            temp.texCoords[i] = Integer.parseInt(st2.nextToken()) - 1;
                            temp.normals[i] = Integer.parseInt(st2.nextToken()) - 1;
                        } else {
                            System.out.println("Too many tokens..");
                            System.exit(1);
                        }
                    }
                    faces.add(temp);
                }
            }
            groups.lastElement().end = faces.size();
            
            for (int i = 0; i < groups.size(); i++) {
                int start = 0;
                int end = 0;

                if (hasTexCoords) {
                    start = vboWt.getCurrIndex();

                    for (int j = groups.elementAt(i).start; j < groups.elementAt(i).end; j++) {
                        vboWt.addTriangle(vertices.elementAt(faces.elementAt(j).vertices[0]), texCoords.elementAt(faces.elementAt(j).texCoords[0]), normals.elementAt(faces.elementAt(j).normals[0]),
                                vertices.elementAt(faces.elementAt(j).vertices[1]), texCoords.elementAt(faces.elementAt(j).texCoords[1]), normals.elementAt(faces.elementAt(j).normals[1]),
                                vertices.elementAt(faces.elementAt(j).vertices[2]), texCoords.elementAt(faces.elementAt(j).texCoords[2]), normals.elementAt(faces.elementAt(j).normals[2]));
                    }


                    end = vboWt.getCurrIndex();
                } else {
                    start = vbo.getCurrIndex();

                    for (int j = groups.elementAt(i).start; j < groups.elementAt(i).end; j++) {
                        vbo.addTriangle(vertices.elementAt(faces.elementAt(j).vertices[0]), normals.elementAt(faces.elementAt(j).normals[0]),
                                vertices.elementAt(faces.elementAt(j).vertices[1]), normals.elementAt(faces.elementAt(j).normals[1]),
                                vertices.elementAt(faces.elementAt(j).vertices[2]), normals.elementAt(faces.elementAt(j).normals[2]));
                    }

                    end = vbo.getCurrIndex();
                }
            }


            CheckError(gl, "ObjModel.loadModel");

        } catch (FileNotFoundException ex) {
            System.out.println("OBJModel: Could not load resource: " + filename);
            throw new FileNotFoundException();
        } catch (Exception e) {
            System.out.println(e.getMessage());
            e.printStackTrace();
            System.exit(1);
        }
    }

    public void draw(GL10 gl)
    {
        if(hasTexCoords)
        {
            vboWt.Draw(gl, 0, numFaces);
        }
        else
        {
           vbo.Draw(gl, 0, numFaces); 
        }
    }
}