/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package de.chaubold.math;

import de.chaubold.math.Vec4;
import java.security.InvalidParameterException;

/**
 *
 * @author caste
 */
public class Matrix4x4 {

    private float values[];

    public Matrix4x4() {
        values = new float[16];
    }

    public Matrix4x4(Matrix4x4 a) {
        for (int i = 0; i < 16; i++) {
            values[i] = a.values[i];
        }
    }

    public Matrix4x4(float[] a) {
        if (a.length != 16) {
            throw new InvalidParameterException("Need a 16 float array!!");
        }
        for (int i = 0; i < 16; i++) {
            values[i] = a[i];
        }
    }

    public float get(int row, int col) {
        return values[row + 4 * col];
    }

    public void set(int row, int col, float value) {
        values[row + 4 * col] = value;
    }

    private void swap(int r1, int c1, int r2, int c2) {
        float tmp = get(r1, c1);
        set(r1, c1, get(r2, c2));
        set(r2, c2, tmp);
    }

    private void gaussj(Matrix4x4 a, Matrix4x4 b) {
        int i, icol = 0, irow = 0, j, k, l, ll, n = 4, m = 4;
        float big, dum, pivinv;
        int[] indxc = new int[n];
        int[] indxr = new int[n];
        int[] ipiv = new int[n];

        for (j = 0; j < n; j++) {
            ipiv[j] = 0;
        }

        for (i = 0; i < n; i++) {
            big = 0.0f;
            for (j = 0; j < n; j++) {
                if (ipiv[j] != 1) {
                    for (k = 0; k < n; k++) {
                        if (ipiv[k] == 0) {
                            if (Math.abs(a.get(j, k)) >= big) {
                                big = Math.abs(a.get(j, k));
                                irow = j;
                                icol = k;
                            }
                        }
                    }
                }
            }
            ++(ipiv[icol]);
            if (irow != icol) {
                for (l = 0; l < n; l++) {
                    a.swap(irow, l, icol, l);
                }
                for (l = 0; l < m; l++) {
                    b.swap(irow, l, icol, l);
                }
            }
            indxr[i] = irow;
            indxc[i] = icol;
            if (a.get(icol, icol) == 0.0) {
                throw new InvalidParameterException("Cannot invert singular matrix");
            }
            pivinv = 1.0f / a.get(icol, icol);
            a.set(icol, icol, 1.0f);
            for (l = 0; l < n; l++) {
                a.set(icol, l, a.get(icol, l) * pivinv);
            }
            for (l = 0; l < m; l++) {
                b.set(icol, l, b.get(icol, l) * pivinv);
            }

            for (ll = 0; ll < n; ll++) {
                if (ll != icol) {
                    dum = a.get(ll, icol);
                    a.set(ll, icol, 0.0f);
                    for (l = 0; l < n; l++) {
                        a.set(ll, l, a.get(ll, l) - a.get(icol, l) * dum);
                    }
                    for (l = 0; l < m; l++) {
                        b.set(ll, l, a.get(ll, l) - b.get(icol, l) * dum);
                    }
                }
            }
        }
//    This is the end of the main loop over columns of the reduction. It only remains to unscram-
//    ble the solution in view of the column interchanges. We do this by interchanging pairs of
//    columns in the reverse order that the permutation was built up.
        for (l = n - 1; l >= 0; l--) {
            if (indxr[l] != indxc[l]) {
                for (k = 0; k < n; k++) {
                    a.swap(k, indxr[l], k, indxc[l]);
                }
            }
        }
    }

    /**
     * Returns the inverse of this matrix
     * @return the inverse
     */
    public Matrix4x4 Inverse() {
        Matrix4x4 tmp = new Matrix4x4(this);
        gaussj(tmp, new Matrix4x4());
        return tmp;
    }


//Basic transformations
    public static Matrix4x4 CreateIdentity() {
        Matrix4x4 mat = new Matrix4x4();
        mat.values[0] = mat.values[5] = mat.values[10] = mat.values[15] = 1.0f;
        return mat;
    }

    public static Matrix4x4 CreateRotationAxis(float degrees, float axis_x, float axis_y, float axis_z) {
        float radians = degrees * (float) Math.PI / 180.0f;
        float rcos = (float) Math.cos(radians);
        float rsin = (float) Math.sin(radians);

        Matrix4x4 out = new Matrix4x4();

        out.values[0] = rcos + axis_x * axis_x * (1 - rcos);
        out.values[1] = axis_z * rsin + axis_y * axis_x * (1 - rcos);
        out.values[2] = -axis_y * rsin + axis_z * axis_x * (1 - rcos);
        out.values[3] = 0.0f;

        out.values[4] = -axis_z * rsin + axis_x * axis_y * (1 - rcos);
        out.values[5] = rcos + axis_y * axis_y * (1 - rcos);
        out.values[6] = axis_x * rsin + axis_z * axis_y * (1 - rcos);
        out.values[7] = 0.0f;

        out.values[8] = axis_y * rsin + axis_x * axis_z * (1 - rcos);
        out.values[9] = -axis_x * rsin + axis_y * axis_z * (1 - rcos);
        out.values[10] = rcos + axis_z * axis_z * (1 - rcos);
        out.values[11] = 0.0f;

        out.values[12] = 0.0f;
        out.values[13] = 0.0f;
        out.values[14] = 0.0f;
        out.values[15] = 1.0f;

        return out;
    }

    public static Matrix4x4 CreateRotationPitchYawRoll(float pitch, float yaw, float roll) {
        Matrix4x4 out = new Matrix4x4();
        double cr = Math.cos(pitch);
        double sr = Math.sin(pitch);
        double cp = Math.cos(yaw);
        double sp = Math.sin(yaw);
        double cy = Math.cos(roll);
        double sy = Math.sin(roll);
        double srsp = sr * sp;
        double crsp = cr * sp;

        out.values[0] = (float) (cp * cy);
        out.values[4] = (float) (cp * sy);
        out.values[8] = (float) -sp;

        out.values[1] = (float) (srsp * cy - cr * sy);
        out.values[5] = (float) (srsp * sy + cr * cy);
        out.values[9] = (float) (sr * cp);

        out.values[2] = (float) (crsp * cy + sr * sy);
        out.values[6] = (float) (crsp * sy - sr * cy);
        out.values[10] = (float) (cr * cp);

        out.values[3] = out.values[7] = out.values[11] = 0.0f;
        out.values[15] = 1.0f;

        return out;
    }

    public static Matrix4x4 CreateScale(float x, float y, float z) {
        Matrix4x4 out = new Matrix4x4();

        out.values[0] = x;
        out.values[5] = y;
        out.values[10] = z;
        out.values[15] = 1.0f;

        return out;
    }

    public static Matrix4x4 CreateTranslation(float x, float y, float z) {
        Matrix4x4 out = new Matrix4x4();

        out.values[12] = x;
        out.values[13] = y;
        out.values[14] = z;
        out.values[15] = 1.0f;

        return out;
    }


//Projection matrices
    public static Matrix4x4 CreatePerspectiveProjection(float fovY, float aspect, float zNear, float zFar) {
        float r = fovY * (float) Math.PI / 360.0f;
        float deltaZ = zFar - zNear;
        float s = (float) Math.sin(r);
        float cotangent = 0;

        if (deltaZ == 0 || s == 0 || aspect == 0) {
            return null;
        }

        //cos(r) / sin(r) = cot(r)
        cotangent = (float) Math.cos(r) / s;

        Matrix4x4 out = new Matrix4x4();

        out.values[0] = cotangent / aspect;
        out.values[5] = cotangent;
        out.values[10] = -(zFar + zNear) / deltaZ;
        out.values[11] = -1;
        out.values[14] = -2 * zNear * zFar / deltaZ;
        out.values[15] = 0;

        return out;
    }

    /** Creates an orthographic projection matrix like glOrtho */
    public static Matrix4x4 CreateOrthographicProjection(float left, float right, float bottom, float top, float nearVal, float farVal) {
        Matrix4x4 out = new Matrix4x4();
        float tx = -((right + left) / (right - left));
        float ty = -((top + bottom) / (top - bottom));
        float tz = -((farVal + nearVal) / (farVal - nearVal));

        out.values[0] = 2 / (right - left);
        out.values[5] = 2 / (top - bottom);
        out.values[10] = -2 / (farVal - nearVal);
        out.values[12] = tx;
        out.values[13] = ty;
        out.values[14] = tz;

        return out;
    }

    /**
     * Builds a translation matrix in the same way as gluLookAt()
     */

//Camera Setup
    public static Matrix4x4 CreateLookAt(Vec3 pEye, Vec3 pCenter, Vec3 pUp) {
        Vec3 f, up, s, u;
        Matrix4x4 out = new Matrix4x4();
        Matrix4x4 translate = new Matrix4x4();
        f = (pCenter.sub(pEye)).normalize();

        up = pUp.normalize();

        s = (Vec3.cross(f, up)).normalize();

        u = (Vec3.cross(s, f)).normalize();

        out.values[0] = s.x;
        out.values[4] = s.y;
        out.values[8] = s.z;

        out.values[1] = u.x;
        out.values[5] = u.y;
        out.values[9] = u.z;

        out.values[2] = -f.x;
        out.values[6] = -f.y;
        out.values[10] = -f.z;

        translate = Matrix4x4.CreateTranslation(-pEye.x, -pEye.y, -pEye.z);
        out = out.Multiply(translate);

        return out;
    }

    public Matrix4x4 transpose() {
        int x, z;
        Matrix4x4 out = new Matrix4x4();

        for (z = 0; z < 4; ++z) {
            for (x = 0; x < 4; ++x) {
                out.values[(z * 4) + x] = values[(x * 4) + z];
            }
        }

        return out;
    }

    public Matrix4x4 Multiply(Matrix4x4 other) {
        Matrix4x4 out = new Matrix4x4();

        out.values[0] = values[0] * other.values[0] + values[4] * other.values[1] + values[8] * other.values[2] + values[12] * other.values[3];
        out.values[1] = values[1] * other.values[0] + values[5] * other.values[1] + values[9] * other.values[2] + values[13] * other.values[3];
        out.values[2] = values[2] * other.values[0] + values[6] * other.values[1] + values[10] * other.values[2] + values[14] * other.values[3];
        out.values[3] = values[3] * other.values[0] + values[7] * other.values[1] + values[11] * other.values[2] + values[15] * other.values[3];

        out.values[4] = values[0] * other.values[4] + values[4] * other.values[5] + values[8] * other.values[6] + values[12] * other.values[7];
        out.values[5] = values[1] * other.values[4] + values[5] * other.values[5] + values[9] * other.values[6] + values[13] * other.values[7];
        out.values[6] = values[2] * other.values[4] + values[6] * other.values[5] + values[10] * other.values[6] + values[14] * other.values[7];
        out.values[7] = values[3] * other.values[4] + values[7] * other.values[5] + values[11] * other.values[6] + values[15] * other.values[7];

        out.values[8] = values[0] * other.values[8] + values[4] * other.values[9] + values[8] * other.values[10] + values[12] * other.values[11];
        out.values[9] = values[1] * other.values[8] + values[5] * other.values[9] + values[9] * other.values[10] + values[13] * other.values[11];
        out.values[10] = values[2] * other.values[8] + values[6] * other.values[9] + values[10] * other.values[10] + values[14] * other.values[11];
        out.values[11] = values[3] * other.values[8] + values[7] * other.values[9] + values[11] * other.values[10] + values[15] * other.values[11];

        out.values[12] = values[0] * other.values[12] + values[4] * other.values[13] + values[8] * other.values[14] + values[12] * other.values[15];
        out.values[13] = values[1] * other.values[12] + values[5] * other.values[13] + values[9] * other.values[14] + values[13] * other.values[15];
        out.values[14] = values[2] * other.values[12] + values[6] * other.values[13] + values[10] * other.values[14] + values[14] * other.values[15];
        out.values[15] = values[3] * other.values[12] + values[7] * other.values[13] + values[11] * other.values[14] + values[15] * other.values[15];

        return out;
    }

    Vec3 getUpVec3() {
        Vec3 out = new Vec3();
        out.x = values[4];
        out.y = values[5];
        out.z = values[6];

        return out;
    }

    Vec3 getRightVec3() {
        Vec3 out = new Vec3();
        out.x = values[0];
        out.y = values[1];
        out.z = values[2];

        return out;
    }

    Vec3 getForwardVec3() {
        Vec3 out = new Vec3();
        out.x = values[8];
        out.y = values[9];
        out.z = values[10];

        return out;
    }

    Vec3 Multiply(Vec3 v) {
        Vec3 out = new Vec3();
        out.x = values[0] * v.x + values[4] * v.y + values[8] * v.z;
        out.y = values[1] * v.x + values[5] * v.y + values[9] * v.z;
        out.z = values[2] * v.x + values[6] * v.y + values[10] * v.z;
        return out;
    }

    Vec4 Multiply(Vec4 v)//returns this * v
    {
        Vec4 out = new Vec4();
        out.x = values[0] * v.x + values[4] * v.y + values[ 8] * v.z + values[12] * v.w;
        out.y = values[1] * v.x + values[5] * v.y + values[ 9] * v.z + values[13] * v.w;
        out.z = values[2] * v.x + values[6] * v.y + values[10] * v.z + values[14] * v.w;
        out.w = values[3] * v.x + values[7] * v.y + values[11] * v.z + values[15] * v.w;
        return out;
    }
}
